import json

from dagster_graphql.test.utils import execute_dagster_graphql

from .utils import define_context

try:
    # Python 2 tempfile doesn't have tempfile.TemporaryDirectory
    import backports.tempfile as tempfile
except ImportError:
    import tempfile


CREATE_PIPELINE_SCHEDULE_QUERY = '''
mutation CreatePipelineScheduleResult($schedule: RunScheduleInput!) {
  createRunSchedule(schedule: $schedule) {
    schedule {
      scheduleId
      name
      cronSchedule
      executionParamsString
    }
  }
}
'''

GET_SCHEDULES_QUERY = '''
{
	scheduler {
        schedulerType
        schedules {
            scheduleId
        }
    }
}
'''

DELETE_SCHEDULE_QUERY = '''
mutation DeletePipelineSchedule($scheduleId: String!) {
  deleteRunSchedule(scheduleId: $scheduleId) {
    deletedSchedule {
      scheduleId
    }
  }
}
'''


def create_default_schedule(context):
    execution_params = {
        "environmentConfigData": {"storage": {"filesystem": None}},
        "selector": {"name": "many_events", "solidSubset": None},
        "mode": "default",
    }

    return create_schedule(context, "run_many_events", "* * * * *", execution_params)


def create_schedule(context, name, cron_schedule, execution_params):
    variables = {
        "schedule": {
            "name": name,
            "cronSchedule": cron_schedule,
            "executionParams": execution_params,
        }
    }

    return execute_dagster_graphql(context, CREATE_PIPELINE_SCHEDULE_QUERY, variables=variables)


def test_create_schedule():
    with tempfile.TemporaryDirectory() as schedule_dir:
        context = define_context(schedule_dir=schedule_dir)

        execution_params = {
            "environmentConfigData": {"storage": {"filesystem": None}},
            "selector": {"name": "many_events", "solidSubset": None},
            "mode": "default",
        }

        result = create_schedule(context, "run_many_events", "* * * * *", execution_params)

        assert result.data
        assert result.data['createRunSchedule']

        schedule = result.data['createRunSchedule']['schedule']
        assert schedule
        assert schedule['scheduleId']
        assert schedule['name'] == "run_many_events"
        assert schedule['cronSchedule'] == "* * * * *"
        assert json.loads(schedule['executionParamsString']) == execution_params


def test_get_all_schedules():
    with tempfile.TemporaryDirectory() as schedule_dir:
        context = define_context(schedule_dir=schedule_dir)

        result = create_default_schedule(context)
        schedule_id = result.data['createRunSchedule']['schedule']['scheduleId']

        # Query Scheduler + all Schedules
        scheduler_result = execute_dagster_graphql(
            context, GET_SCHEDULES_QUERY, variables={"pipelineName": "many_events"}
        )

        assert scheduler_result.data
        assert scheduler_result.data['scheduler']
        assert scheduler_result.data['scheduler']['schedules']
        assert len(scheduler_result.data['scheduler']['schedules']) == 1

        assert scheduler_result.data['scheduler']['schedules'][0]['scheduleId'] == schedule_id


def test_delete_schedule():
    with tempfile.TemporaryDirectory() as schedule_dir:
        context = define_context(schedule_dir=schedule_dir)

        result = create_default_schedule(context)
        schedule_id = result.data['createRunSchedule']['schedule']['scheduleId']

        # Delete Schedule
        delete_result = execute_dagster_graphql(
            context, DELETE_SCHEDULE_QUERY, variables={"scheduleId": schedule_id}
        )

        assert delete_result.data
        assert delete_result.data['deleteRunSchedule']
        assert delete_result.data['deleteRunSchedule']['deletedSchedule']
        assert (
            delete_result.data['deleteRunSchedule']['deletedSchedule']['scheduleId'] == schedule_id
        )

        # Query all Schedules
        scheduler_result = execute_dagster_graphql(
            context, GET_SCHEDULES_QUERY, variables={"pipelineName": "many_events"}
        )

        assert scheduler_result.data
        assert len(scheduler_result.data['scheduler']['schedules']) == 0



import json

from dagster import check
from dagster_graphql import dauphin

from dagster_graphql.implementation.scheduler import RunSchedule, SystemCronScheduler


class SchedulerType(dauphin.Enum):
    SystemCronScheduler = "SystemCronScheduler"


def get_scheduler(graphene_info):
    scheduler = graphene_info.context.scheduler

    if isinstance(scheduler, SystemCronScheduler):
        scheduler_type = SchedulerType.SystemCronScheduler
    else:
        raise Exception(
            'Unknown Scheduler type {typ}. Add this Scheduler type to the SchedulerType '
            'GraphQL Enum'.format(typ=type(scheduler))
        )

    schedules = [
        graphene_info.schema.type_named('RunSchedule')(s) for s in scheduler.all_schedules()
    ]

    return graphene_info.schema.type_named('Scheduler')(
        scheduler_type=scheduler_type, schedules=schedules
    )


class DauphinScheduler(dauphin.ObjectType):
    class Meta:
        name = 'Scheduler'

    scheduler_type = dauphin.NonNull(dauphin.String)
    schedules = dauphin.non_null_list('RunSchedule')


class DauphinRunScheduleInput(dauphin.InputObjectType):
    class Meta:
        name = 'RunScheduleInput'

    name = dauphin.NonNull(dauphin.String)
    cron_schedule = dauphin.NonNull(dauphin.String)
    execution_params = dauphin.NonNull('ExecutionParams')


class DauphinCreateRunScheduleResult(dauphin.ObjectType):
    class Meta:
        name = 'CreateRunScheduleResult'

    schedule = dauphin.NonNull('RunSchedule')


class DauphinCreateRunScheduleMutation(dauphin.Mutation):
    class Meta:
        name = 'CreateRunScheduleMutation'

    class Arguments:
        schedule = dauphin.NonNull('RunScheduleInput')

    Output = dauphin.NonNull('CreateRunScheduleResult')

    def mutate(self, graphene_info, **kwargs):
        from dagster_graphql.schema.roots import create_execution_params

        # Check execution_params is valid ExecutionParams
        create_execution_params(kwargs['schedule'].get('execution_params'))

        scheduler = graphene_info.context.scheduler
        schedule = scheduler.create_schedule(
            name=kwargs['schedule'].get('name'),
            cron_schedule=kwargs['schedule'].get('cron_schedule'),
            execution_params=kwargs['schedule'].get('execution_params'),
        )

        return graphene_info.schema.type_named('CreateRunScheduleResult')(
            schedule=graphene_info.schema.type_named('RunSchedule')(schedule)
        )


class DauphinDeleteRunScheduleResult(dauphin.ObjectType):
    class Meta:
        name = 'DeleteRunScheduleResult'

    deleted_schedule = dauphin.NonNull('RunSchedule')


class DauphineDeleteRunScheduleMutation(dauphin.Mutation):
    class Meta:
        name = 'DeleteRunScheduleMutation'

    class Arguments:
        schedule_id = dauphin.NonNull(dauphin.String)

    Output = dauphin.NonNull('DeleteRunScheduleResult')

    def mutate(self, graphene_info, **kwargs):
        scheduler = graphene_info.context.scheduler
        schedule = scheduler.remove_schedule(kwargs['schedule_id'])

        return graphene_info.schema.type_named('DeleteRunScheduleResult')(
            deleted_schedule=graphene_info.schema.type_named('RunSchedule')(schedule)
        )


class DauphinRunSchedule(dauphin.ObjectType):
    class Meta:
        name = 'RunSchedule'

    schedule_id = dauphin.NonNull(dauphin.String)
    name = dauphin.NonNull(dauphin.String)
    cron_schedule = dauphin.NonNull(dauphin.String)
    execution_params_string = dauphin.NonNull(dauphin.String)

    def __init__(self, schedule):
        self._schedule = check.inst_param(schedule, 'schedule', RunSchedule)
        super(DauphinRunSchedule, self).__init__(
            schedule_id=schedule.schedule_id,
            name=schedule.name,
            cron_schedule=schedule.cron_schedule,
            execution_params_string=json.dumps(schedule.execution_params),
        )



import six
import abc
from collections import namedtuple

from dagster import check


class Scheduler(six.with_metaclass(abc.ABCMeta)):
    @abc.abstractmethod
    def create_schedule(self, *args, **kwargs):
        '''Create a new pipeline schedule.

        Passes args and kwargs to the new Schedule.

        Returns:
            PipelineSchedule
        '''

    @abc.abstractmethod
    def remove_schedule(self, id_):
        '''Deletes a pipeline schedule.

        Args:
            id_ (str): The id of the schedule
        '''

    @abc.abstractmethod
    def all_schedules(self):
        '''Return all the schedules present in the storage.

        Returns:
            Iterable[(str, PipelineSchedule)]: Tuples of schedule_id, schedule.
        '''

    @abc.abstractmethod
    def all_schedules_for_pipeline(self, pipeline_name):
        '''Return all the schedules present in the storage for a given pipeline.

        Args:
            pipeline_name (str): The pipeline to index on

        Returns:
            Iterable[(str, PipelineSchedule)]: Tuples of schedule_id, schedule.
        '''

    @abc.abstractmethod
    def get_schedule_by_id(self, id_):
        '''Get a schedule by its id.

        Args:
            id_ (str): The id of the schedule

        Returns:
            Optional[PipelineSchedule]
        '''

    @abc.abstractmethod
    def start_schedule(self, id_):
        '''Start a pipeline schedule.

        Args:
            id_ (str): The id of the schedule
        '''

    @abc.abstractmethod
    def end_schedule(self, id_):
        '''Ends an existing pipeline schedule

        Args:
            id_ (str): The id of the schedule
        '''


class RunSchedule(namedtuple('RunSchedule', 'schedule_id name cron_schedule execution_params')):
    def __new__(cls, schedule_id, name, cron_schedule, execution_params):
        return super(RunSchedule, cls).__new__(
            cls,
            check.str_param(schedule_id, 'schedule_id'),
            check.str_param(name, 'name'),
            check.str_param(cron_schedule, 'cron_schedule'),
            check.dict_param(execution_params, 'execution_params'),
        )



import io
import uuid

import json
import os

from collections import OrderedDict

import six

from dagster import check, seven, utils

from .scheduler import Scheduler, RunSchedule


class SystemCronScheduler(Scheduler):
    def __init__(self, schedule_dir):
        check.str_param(schedule_dir, 'schedule_dir')
        self._schedule_dir = schedule_dir

        self._schedules = OrderedDict()
        self._load_schedules()

    def create_schedule(self, *args, **kwargs):
        schedule_id = str(uuid.uuid4())
        schedule = RunSchedule(schedule_id=schedule_id, *args, **kwargs)
        self._write_schedule_to_file(schedule)
        self._schedules[schedule_id] = schedule
        return schedule

    def remove_schedule(self, id_):
        schedule = self._schedules.pop(id_)
        metadata_file = self._get_metadata_file_path(schedule)
        os.remove(metadata_file)
        return schedule

    def all_schedules(self):
        return [s for s in self._schedules.values()]

    def all_schedules_for_pipeline(self, pipeline_name):
        return [
            s
            for s in self.all_schedules()
            if s.execution_params['selector']['name'] == pipeline_name
        ]

    def get_schedule_by_id(self, id_):
        return self._schedules.get(id_)

    def start_schedule(self, id_):
        # TODO: Create cron job and save to user's crontab
        check.not_implemented("not implemented")

    def end_schedule(self, id_):
        # TODO: Remove cron job from user's crontab
        check.not_implemented("not implmented")

    def _get_file_prefix(self, schedule):
        return os.path.join(self._schedule_dir, '{}_{}'.format(schedule.name, schedule.schedule_id))

    def _get_metadata_file_path(self, schedule):
        file_prefix = self._get_file_prefix(schedule)
        return '{}.json'.format(file_prefix)

    def _write_schedule_to_file(self, schedule):
        metadata_file = self._get_metadata_file_path(schedule)
        with io.open(metadata_file, 'w', encoding='utf-8') as f:
            json_str = seven.json.dumps(
                {
                    'schedule_id': schedule.schedule_id,
                    'name': schedule.name,
                    'cron_schedule': schedule.cron_schedule,
                    'execution_params': schedule.execution_params,
                }
            )
            f.write(six.text_type(json_str))

        return metadata_file

    def _load_schedules(self):
        utils.mkdir_p(self._schedule_dir)

        for file in os.listdir(self._schedule_dir):
            if not file.endswith('.json'):
                continue
            file_path = os.path.join(self._schedule_dir, file)
            with open(file_path) as data:
                try:
                    data = json.load(data)
                    schedule = RunSchedule(
                        schedule_id=data['schedule_id'],
                        name=data['name'],
                        cron_schedule=data['cron_schedule'],
                        execution_params=data['execution_params'],
                    )
                    self._schedules[schedule.schedule_id] = schedule

                except Exception as ex:  # pylint: disable=broad-except
                    six.raise_from(
                        Exception(
                            'Could not parse dagit schedule from {file_name} in {dir_name}. {ex}: {msg}'.format(
                                file_name=file,
                                dir_name=self._schedule_dir,
                                ex=type(ex).__name__,
                                msg=ex,
                            )
                        ),
                        ex,
                    )
